@@ -21,6 +21,7 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <stdarg.h>
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -73,6 +74,7 @@ win_t win;
 fileinfo_t *files;
 int filecnt, fileidx;
 int alternate;
+int markcnt;
 
 int prefix;
 bool extprefix;
@@ -111,7 +113,7 @@ void cleanup(void)
 	}
 }
 
-void check_add_file(char *filename)
+void check_add_file(char *filename, bool given)
 {
 	const char *bn;
 
@@ -119,7 +121,8 @@ void check_add_file(char *filename)
 		return;
 
 	if (access(filename, R_OK) < 0) {
-		warn("could not open file: %s", filename);
+		if (given)
+			warn("could not open file: %s", filename);
 		return;
 	}
 
@@ -146,7 +149,7 @@ void check_add_file(char *filename)
 	}
 #endif
 
-	files[fileidx].loaded = false;
+	files[fileidx].warn = given;
 	files[fileidx].name = s_strdup(filename);
 	if (files[fileidx].path == NULL)
 		files[fileidx].path = files[fileidx].name;
@@ -168,22 +171,24 @@ void remove_file(int n, bool manual)
 		cleanup();
 		exit(manual ? EXIT_SUCCESS : EXIT_FAILURE);
 	}
+	if (files[n].marked)
+		markcnt--;
 
 	if (files[n].path != files[n].name)
 		free((void*) files[n].path);
 	free((void*) files[n].name);
 
-	if (n + 1 < filecnt)
-		memmove(files + n, files + n + 1, (filecnt - n - 1) * sizeof(fileinfo_t));
-	if (n + 1 < tns.cnt) {
-		memmove(tns.thumbs + n, tns.thumbs + n + 1, (tns.cnt - n - 1) *
-		        sizeof(thumb_t));
-		memset(tns.thumbs + tns.cnt - 1, 0, sizeof(thumb_t));
+	if (n + 1 < filecnt) {
+		if (tns.thumbs != NULL) {
+			memmove(tns.thumbs + n, tns.thumbs + n + 1, (filecnt - n - 1) *
+			        sizeof(*tns.thumbs));
+			memset(tns.thumbs + filecnt - 1, 0, sizeof(*tns.thumbs));
+		}
+		memmove(files + n, files + n + 1, (filecnt - n - 1) * sizeof(*files));
 	}
-
 	filecnt--;
-	if (n < tns.cnt)
-		tns.cnt--;
+	if (fileidx >= filecnt)
+		fileidx = filecnt - 1;
 	if (n < alternate)
 		alternate--;
 }
@@ -253,7 +258,7 @@ void open_info(void)
 		kill(pid, SIGTERM);
 		info.fd = -1;
 	}
-	win.bar.l[0] = '\0';
+	win.bar.l.buf[0] = '\0';
 
 	if (pipe(pfd) < 0)
 		return;
@@ -287,21 +292,21 @@ void read_info(void)
 		for (i = 0; i < n; i++) {
 			if (buf[i] == '\n') {
 				if (info.lastsep == 0) {
-					win.bar.l[info.i++] = ' ';
+					win.bar.l.buf[info.i++] = ' ';
 					info.lastsep = 1;
 				}
 			} else {
-				win.bar.l[info.i++] = buf[i];
+				win.bar.l.buf[info.i++] = buf[i];
 				info.lastsep = 0;
 			}
-			if (info.i + 1 == sizeof(win.bar.l))
+			if (info.i + 1 == win.bar.l.size)
 				goto end;
 		}
 	}
 end:
 	info.i -= info.lastsep;
-	win.bar.l[info.i] = '\0';
-	win_update_bar(&win);
+	win.bar.l.buf[info.i] = '\0';
+	win_draw(&win);
 	close(info.fd);
 	info.fd = -1;
 	while (waitpid(-1, NULL, WNOHANG) > 0);
@@ -309,14 +314,16 @@ end:
 
 void load_image(int new)
 {
+	static int current;
+
 	if (new < 0 || new >= filecnt)
 		return;
 
 	win_set_cursor(&win, CURSOR_WATCH);
 	reset_timeout(slideshow);
 
-	if (new != fileidx)
-		alternate = fileidx;
+	if (new != current)
+		alternate = current;
 
 	img_close(&img, false);
 	while (!img_load(&img, &files[new])) {
@@ -326,8 +333,8 @@ void load_image(int new)
 		else if (new > 0 && new < fileidx)
 			new--;
 	}
-	files[new].loaded = true;
-	fileidx = new;
+	files[new].warn = false;
+	fileidx = current = new;
 
 	info.open = false;
 	open_info();
@@ -338,63 +345,71 @@ void load_image(int new)
 		reset_timeout(animate);
 }
 
+void bar_put(win_bar_t *bar, const char *fmt, ...)
+{
+	size_t len = bar->size - (bar->p - bar->buf), n;
+	va_list ap;
+
+	va_start(ap, fmt);
+	n = vsnprintf(bar->p, len, fmt, ap);
+	bar->p += MIN(len, n);
+	va_end(ap);
+}
+
 void update_info(void)
 {
-	int sel;
-	unsigned int i, fn, fw, n;
-	unsigned int llen = sizeof(win.bar.l), rlen = sizeof(win.bar.r);
-	char *lt = win.bar.l, *rt = win.bar.r, title[TITLE_LEN];
+	unsigned int i, fn, fw;
+	char title[TITLE_LEN];
 	const char * mark;
 	bool ow_info;
-
-	for (fw = 0, i = filecnt; i > 0; fw++, i /= 10);
-	sel = mode == MODE_IMAGE ? fileidx : tns.sel;
+	win_bar_t *l = &win.bar.l, *r = &win.bar.r;
 
 	/* update window title */
 	if (mode == MODE_THUMB) {
 		win_set_title(&win, "sxiv");
 	} else {
-		snprintf(title, sizeof(title), "sxiv - %s", files[sel].name);
+		snprintf(title, sizeof(title), "sxiv - %s", files[fileidx].name);
 		win_set_title(&win, title);
 	}
 
 	/* update bar contents */
 	if (win.bar.h == 0)
 		return;
-	mark = files[sel].marked ? "* " : "";
+	for (fw = 0, i = filecnt; i > 0; fw++, i /= 10);
+	mark = files[fileidx].marked ? "* " : "";
+	l->p = l->buf;
+	r->p = r->buf;
 	if (mode == MODE_THUMB) {
-		if (tns.cnt == filecnt) {
-			n = snprintf(rt, rlen, "%s%0*d/%d", mark, fw, sel + 1, filecnt);
-			ow_info = true;
-		} else {
-			snprintf(lt, llen, "Loading... %0*d/%d", fw, tns.cnt, filecnt);
-			rt[0] = '\0';
+		if (tns.loadnext < tns.end) {
+			bar_put(l, "Loading... %0*d", fw, MAX(tns.loadnext, 1));
 			ow_info = false;
+		} else {
+			ow_info = true;
 		}
+		bar_put(r, "%s%0*d/%d", mark, fw, fileidx + 1, filecnt);
 	} else {
-		n = snprintf(rt, rlen, "%s", mark);
+		bar_put(r, "%s", mark);
 		if (img.ss.on)
-			n += snprintf(rt + n, rlen - n, "%ds | ", img.ss.delay);
+			bar_put(r, "%ds | ", img.ss.delay);
 		if (img.gamma != 0)
-			n += snprintf(rt + n, rlen - n, "G%+d | ", img.gamma);
-		n += snprintf(rt + n, rlen - n, "%3d%% | ", (int) (img.zoom * 100.0));
+			bar_put(r, "G%+d | ", img.gamma);
+		bar_put(r, "%3d%% | ", (int) (img.zoom * 100.0));
 		if (img.multi.cnt > 0) {
 			for (fn = 0, i = img.multi.cnt; i > 0; fn++, i /= 10);
-			n += snprintf(rt + n, rlen - n, "%0*d/%d | ",
-			              fn, img.multi.sel + 1, img.multi.cnt);
+			bar_put(r, "%0*d/%d | ", fn, img.multi.sel + 1, img.multi.cnt);
 		}
-		n += snprintf(rt + n, rlen - n, "%0*d/%d", fw, sel + 1, filecnt);
+		bar_put(r, "%0*d/%d", fw, fileidx + 1, filecnt);
 		ow_info = info.cmd == NULL;
 	}
 	if (ow_info) {
-		fn = strlen(files[sel].name);
-		if (fn < llen &&
-		    win_textwidth(files[sel].name, fn, true) +
-		    win_textwidth(rt, n, true) < win.w)
+		fn = strlen(files[fileidx].name);
+		if (fn < l->size &&
+		    win_textwidth(files[fileidx].name, fn, true) +
+		    win_textwidth(r->buf, r->p - r->buf, true) < win.w)
 		{
-			strncpy(lt, files[sel].name, llen);
+			strncpy(l->buf, files[fileidx].name, l->size);
 		} else {
-			strncpy(lt, files[sel].base, llen);
+			strncpy(l->buf, files[fileidx].base, l->size);
 		}
 	}
 }
@@ -434,7 +449,7 @@ void reset_cursor(void)
 			}
 		}
 	} else {
-		if (tns.cnt != filecnt)
+		if (tns.loadnext < tns.end)
 			cursor = CURSOR_WATCH;
 		else
 			cursor = CURSOR_ARROW;
@@ -444,7 +459,7 @@ void reset_cursor(void)
 
 void animate(void)
 {
-	if (img_frame_animate(&img, false)) {
+	if (img_frame_animate(&img)) {
 		redraw();
 		set_timeout(animate, img.multi.frames[img.multi.sel].delay, true);
 	}
@@ -464,9 +479,12 @@ void clear_resize(void)
 void run_key_handler(const char *key, unsigned int mask)
 {
 	pid_t pid;
-	int retval, status, n = mode == MODE_IMAGE ? fileidx : tns.sel;
-	char kstr[32];
-	struct stat oldst, newst;
+	int i, j, retval, status;
+	int fcnt = mode == MODE_THUMB && markcnt > 0 ? markcnt : 1;
+	bool changed = false;
+	char **args, kstr[32], oldbar[BAR_L_LEN];
+	struct stat *oldst, newst;
+	struct { int fn; struct stat st; } *finfo;
 
 	if (keyhandler.cmd == NULL) {
 		if (!keyhandler.warned) {
@@ -478,49 +496,70 @@ void run_key_handler(const char *key, unsigned int mask)
 	if (key == NULL)
 		return;
 
+	finfo = s_malloc(fcnt * sizeof(*finfo));
+	args = s_malloc((fcnt + 3) * sizeof(*args));
+	args[0] = keyhandler.cmd;
+	args[1] = kstr;
+	args[fcnt+2] = NULL;
+	if (mode == MODE_IMAGE || markcnt == 0) {
+		finfo[0].fn = fileidx;
+		stat(files[fileidx].path, &finfo[0].st);
+		args[2] = (char*) files[fileidx].path;
+	} else for (i = j = 0; i < filecnt; i++) {
+		if (files[i].marked) {
+			finfo[j].fn = i;
+			stat(files[i].path, &finfo[j++].st);
+			args[j+1] = (char*) files[i].path;
+		}
+	}
 	snprintf(kstr, sizeof(kstr), "%s%s%s%s",
 	         mask & ControlMask ? "C-" : "",
 	         mask & Mod1Mask    ? "M-" : "",
 	         mask & ShiftMask   ? "S-" : "", key);
 
-	stat(files[n].path, &oldst);
+	memcpy(oldbar, win.bar.l.buf, sizeof(oldbar));
+	strncpy(win.bar.l.buf, "Running key handler...", win.bar.l.size);
+	win_draw(&win);
+	win_set_cursor(&win, CURSOR_WATCH);
 
 	if ((pid = fork()) == 0) {
-		execl(keyhandler.cmd, keyhandler.cmd, kstr, files[n].path, NULL);
+		execv(keyhandler.cmd, args);
 		warn("could not exec key handler");
 		exit(EXIT_FAILURE);
 	} else if (pid < 0) {
 		warn("could not fork key handler");
-		return;
+		goto end;
 	}
-	win_set_cursor(&win, CURSOR_WATCH);
-
 	waitpid(pid, &status, 0);
 	retval = WEXITSTATUS(status);
 	if (WIFEXITED(status) == 0 || retval != 0)
 		warn("key handler exited with non-zero return value: %d", retval);
 
-	if (stat(files[n].path, &newst) == 0 &&
-	    memcmp(&oldst.st_mtime, &newst.st_mtime, sizeof(oldst.st_mtime)) == 0)
-	{
-		/* file has not changed */
-		win_set_cursor(&win, CURSOR_ARROW);
-		set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
-		return;
+	for (i = 0; i < fcnt; i++) {
+		oldst = &finfo[i].st;
+		if (stat(files[finfo[i].fn].path, &newst) != 0 ||
+				memcmp(&oldst->st_mtime, &newst.st_mtime, sizeof(newst.st_mtime)) != 0)
+		{
+			if (tns.thumbs != NULL) {
+				tns_unload(&tns, finfo[i].fn);
+				tns.loadnext = MIN(tns.loadnext, finfo[i].fn);
+			}
+			changed = true;
+		}
 	}
+end:
 	if (mode == MODE_IMAGE) {
-		img_close(&img, true);
-		load_image(fileidx);
-	}
-	if (!tns_load(&tns, n, &files[n], true, mode == MODE_IMAGE) &&
-	    mode == MODE_THUMB)
-	{
-		remove_file(tns.sel, false);
-		tns.dirty = true;
-		if (tns.sel >= tns.cnt)
-			tns.sel = tns.cnt - 1;
+		if (changed) {
+			img_close(&img, true);
+			load_image(fileidx);
+		} else if (info.cmd != NULL) {
+			memcpy(win.bar.l.buf, oldbar, win.bar.l.size);
+		}
 	}
+	reset_cursor();
 	redraw();
+	free(finfo);
+	free(args);
 }
 
 #define MODMASK(mask) ((mask) & (ShiftMask|ControlMask|Mod1Mask))
@@ -531,6 +570,7 @@ void on_keypress(XKeyEvent *kev)
 	unsigned int sh;
 	KeySym ksym, shksym;
 	char key;
+	bool dirty = false;
 
 	if (kev == NULL)
 		return;
@@ -557,23 +597,22 @@ void on_keypress(XKeyEvent *kev)
 	} else for (i = 0; i < ARRLEN(keys); i++) {
 		if (keys[i].ksym == ksym &&
 		    MODMASK(keys[i].mask | sh) == MODMASK(kev->state) &&
-		    keys[i].cmd != NULL)
+		    keys[i].cmd >= 0 && keys[i].cmd < CMD_COUNT &&
+		    (cmds[keys[i].cmd].mode < 0 || cmds[keys[i].cmd].mode == mode))
 		{
-			cmdreturn_t ret = keys[i].cmd(keys[i].arg);
-
-			if (ret == CMD_INVALID)
-				continue;
-			if (ret == CMD_DIRTY)
-				redraw();
-			break;
+			if (cmds[keys[i].cmd].func(keys[i].arg))
+				dirty = true;
 		}
 	}
+	if (dirty)
+		redraw();
 	prefix = 0;
 }
 
 void on_buttonpress(XButtonEvent *bev)
 {
 	int i, sel;
+	bool dirty = false;
 	static Time firstclick;
 
 	if (bev == NULL)
@@ -586,32 +625,30 @@ void on_buttonpress(XButtonEvent *bev)
 		for (i = 0; i < ARRLEN(buttons); i++) {
 			if (buttons[i].button == bev->button &&
 			    MODMASK(buttons[i].mask) == MODMASK(bev->state) &&
-			    buttons[i].cmd != NULL)
+			    buttons[i].cmd >= 0 && buttons[i].cmd < CMD_COUNT &&
+			    (cmds[buttons[i].cmd].mode < 0 || cmds[buttons[i].cmd].mode == mode))
 			{
-				cmdreturn_t ret = buttons[i].cmd(buttons[i].arg);
-
-				if (ret == CMD_INVALID)
-					continue;
-				if (ret == CMD_DIRTY)
-					redraw();
-				break;
+				if (cmds[buttons[i].cmd].func(buttons[i].arg))
+					dirty = true;
 			}
 		}
+		if (dirty)
+			redraw();
 	} else {
 		/* thumbnail mode (hard-coded) */
 		switch (bev->button) {
 			case Button1:
 				if ((sel = tns_translate(&tns, bev->x, bev->y)) >= 0) {
-					if (sel != tns.sel) {
-						tns_highlight(&tns, tns.sel, false);
+					if (sel != fileidx) {
+						tns_highlight(&tns, fileidx, false);
 						tns_highlight(&tns, sel, true);
-						tns.sel = sel;
+						fileidx = sel;
 						firstclick = bev->time;
 						redraw();
 					} else if (bev->time - firstclick <= TO_DOUBLE_CLICK) {
 						mode = MODE_IMAGE;
 						set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
-						load_image(tns.sel);
+						load_image(fileidx);
 						redraw();
 					} else {
 						firstclick = bev->time;
@@ -621,6 +658,7 @@ void on_buttonpress(XButtonEvent *bev)
 			case Button3:
 				if ((sel = tns_translate(&tns, bev->x, bev->y)) >= 0) {
 					files[sel].marked = !files[sel].marked;
+					markcnt += files[sel].marked ? 1 : -1;
 					tns_mark(&tns, sel, files[sel].marked);
 					redraw();
 				}
@@ -641,44 +679,39 @@ void run(void)
 	int xfd;
 	fd_set fds;
 	struct timeval timeout;
-	bool discard, to_set;
+	bool discard, load_thumb, to_set;
 	XEvent ev, nextev;
 
-	redraw();
-
 	while (true) {
-		while (mode == MODE_THUMB && tns.cnt < filecnt &&
-		       XPending(win.env.dpy) == 0)
-		{
-			/* load thumbnails */
-			set_timeout(redraw, TO_REDRAW_THUMBS, false);
-			if (tns_load(&tns, tns.cnt, &files[tns.cnt], false, false)) {
-				tns.cnt++;
-			} else {
-				remove_file(tns.cnt, false);
-				if (tns.sel > 0 && tns.sel >= tns.cnt)
-					tns.sel--;
-			}
-			if (tns.cnt == filecnt)
-				redraw();
-			else
-				check_timeouts(NULL);
-		}
+		to_set = check_timeouts(&timeout);
+		load_thumb = mode == MODE_THUMB && tns.loadnext < tns.end;
 
-		while (XPending(win.env.dpy) == 0
-		       && ((to_set = check_timeouts(&timeout)) || info.fd != -1))
+		if ((load_thumb || to_set || info.fd != -1) &&
+		    XPending(win.env.dpy) == 0)
 		{
-			/* check for timeouts & input */
-			xfd = ConnectionNumber(win.env.dpy);
-			FD_ZERO(&fds);
-			FD_SET(xfd, &fds);
-			if (info.fd != -1) {
-				FD_SET(info.fd, &fds);
-				xfd = MAX(xfd, info.fd);
+			if (load_thumb) {
+				set_timeout(redraw, TO_REDRAW_THUMBS, false);
+				if (!tns_load(&tns, tns.loadnext, false)) {
+					remove_file(tns.loadnext, false);
+					tns.dirty = true;
+				}
+				while (tns.loadnext < tns.end && tns.thumbs[tns.loadnext].im != NULL)
+					tns.loadnext++;
+				if (tns.loadnext >= tns.end)
+					redraw();
+			} else {
+				xfd = ConnectionNumber(win.env.dpy);
+				FD_ZERO(&fds);
+				FD_SET(xfd, &fds);
+				if (info.fd != -1) {
+					FD_SET(info.fd, &fds);
+					xfd = MAX(xfd, info.fd);
+				}
+				select(xfd + 1, &fds, 0, 0, to_set ? &timeout : NULL);
+				if (info.fd != -1 && FD_ISSET(info.fd, &fds))
+					read_info();
 			}
-			select(xfd + 1, &fds, 0, 0, to_set ? &timeout : NULL);
-			if (info.fd != -1 && FD_ISSET(info.fd, &fds))
-				read_info();
+			continue;
 		}
 
 		do {
@@ -724,9 +757,6 @@ void run(void)
 					}
 				}
 				break;
-			case Expose:
-				win_expose(&win, &ev.xexpose);
-				break;
 			case KeyPress:
 				on_keypress(&ev.xkey);
 				break;
@@ -758,7 +788,7 @@ int main(int argc, char **argv)
 	parse_options(argc, argv);
 
 	if (options->clean_cache) {
-		tns_init(&tns, 0, NULL);
+		tns_init(&tns, NULL, NULL, NULL, NULL);
 		tns_clean_cache(&tns);
 		exit(EXIT_SUCCESS);
 	}
@@ -781,10 +811,9 @@ int main(int argc, char **argv)
 		while ((len = get_line(&filename, &n, stdin)) > 0) {
 			if (filename[len-1] == '\n')
 				filename[len-1] = '\0';
-			check_add_file(filename);
+			check_add_file(filename, true);
 		}
-		if (filename != NULL)
-			free(filename);
+		free(filename);
 	}
 
 	for (i = 0; i < options->filecnt; i++) {
@@ -795,7 +824,7 @@ int main(int argc, char **argv)
 			continue;
 		}
 		if (!S_ISDIR(fstats.st_mode)) {
-			check_add_file(filename);
+			check_add_file(filename, true);
 		} else {
 			if (!options->recursive) {
 				warn("ignoring directory: %s", filename);
@@ -807,7 +836,7 @@ int main(int argc, char **argv)
 			}
 			start = fileidx;
 			while ((filename = r_readdir(&dir)) != NULL) {
-				check_add_file(filename);
+				check_add_file(filename, false);
 				free((void*) filename);
 			}
 			r_closedir(&dir);
@@ -851,17 +880,18 @@ int main(int argc, char **argv)
 
 	if (options->thumb_mode) {
 		mode = MODE_THUMB;
-		tns_init(&tns, filecnt, &win);
-		while (!tns_load(&tns, 0, &files[0], false, false))
+		tns_init(&tns, files, &filecnt, &fileidx, &win);
+		while (!tns_load(&tns, 0, false))
 			remove_file(0, false);
-		tns.cnt = 1;
 	} else {
 		mode = MODE_IMAGE;
 		tns.thumbs = NULL;
 		load_image(fileidx);
 	}
-
 	win_open(&win);
+	win_set_cursor(&win, CURSOR_WATCH);
+
+	set_timeout(redraw, 25, false);
 
 	run();
 	cleanup();
